/**********************************************************************
  Filename    : functions.h
  Description : app used methods and functions
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/09/09
**********************************************************************/

#include "BluetoothSerial.h"
#include "string.h"

#include <Arduino.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Adafruit_ST7789.h> // Hardware-specific library for ST7789

#include <TJpg_Decoder.h>

// for 16 bit data (65k color mode) image frame data conversion
// remote sends 2 bytes for one pixel, and convert into 16 bit integer
// ST7789: pixel size is 16 bit (2 bytes)
union ArrayToInteger16 {
  byte barry[2];
  uint16_t integer16;
};

// for 8 bit data (256 color mode) image frame data conversion
// remote sends 1 byte for one pixel, and convert into 16 bit integer
// ST7789: pixel size if 16 bit (2 bytes)
uint16_t ByteToInteger16(byte pixel) {
  uint16_t int16 = 0;
  
  int red = round((pixel >> 5) * 255 / 7);
  int green = round(((pixel >> 2) & 0x7) * 255 / 7);
  int blue = round((pixel & 0x3) * 255 / 3);

  int r = (red >> 3);
  int g = (green >> 2);
  int b = (blue >> 3);  

  int16 = (r << 11 | g << 5 | b);
  return int16;
}

// send current display type to remote camera
// remote camera device will sets proper image frame data and status information for transmision
void sendDisplayTypeToRemote() {
  if (DISPLAY_COLOR_MODE == 0) {
    SerialBT.println("displaytype:none");
  } else if (DISPLAY_COLOR_MODE == 1) {
    SerialBT.println("displaytype:SD1306");
  } else if (DISPLAY_COLOR_MODE == 2) {
    SerialBT.println("displaytype:ST7789");
  }
  SerialBT.write(13);
  SerialBT.write(10);
}
// send Jpeg decode mode to remote camera
// remote camera device will send preview frame as jpeg format
void sendImageTypeToRemote() {
  if (imageJpeg) {
    SerialBT.println("imagetype:jpeg");
    SerialBT.write(13);
    SerialBT.write(10);    
  }
}

// clear specific area of the display
void clearArea(uint16_t x, uint16_t y, uint16_t w, uint16_t h, uint16_t color) {
  tft.fillRect(x, y, w, h, color);
}

// display bitmap image on specific location and size
void displayBitmap(int16_t x, int16_t y, uint16_t bitmap[], int16_t w, int16_t h) {
  //Draw a PROGMEM-resident 16-bit image (RGB 5/6/5) at the specified (x,y) position. For 16-bit display devices; no color reduction performed.
  tft.drawRGBBitmap(x, y, bitmap, w, h);
}

// display text
void writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString) {
  tft.setTextWrap(true);
  tft.setCursor(x, y);
  tft.setTextSize(textSize);
  tft.setTextColor(textColor);
  tft.println(textString);
}

bool isPairedDevice(uint8_t address[]) {
  //return memcmp(otherAddress.getNative(), m_address, 6) == 0;
  bool paired = false;
  for (int i = 0; i < PAIR_MAX_DEVICES; i++) {
    if (memcmp(address, pairedDeviceAddress[i], 6) == 0) {
      paired = true;
      break;
    }
  }
  return paired;
}

void displayDevices() {

  bool devicesFound = true;
  
  if (!scanResult) {
    devicesFound = false;
  } else {
    devicesCount = 0;
    devicesCount = scanResult->getCount();
    if (devicesCount < 1) {
      devicesFound = false;
    }
  }

// writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString)
  tft.fillScreen(ST77XX_BLACK);

  if (!devicesFound) {
    writeText(20, 100, 2, ST77XX_WHITE, "No device found.");
    return;  
  }

  int x = 10;
  int y = 20;

  if (selectedIndex == 0) {
    tft.fillRect(0, 0, 240, 60, ST77XX_WHITE);
  } else if (selectedIndex == 1) {
    tft.fillRect(0, 60, 240, 60, ST77XX_WHITE);
  } else if (selectedIndex == 2) {
    tft.fillRect(0, 120, 240, 60, ST77XX_WHITE);
  } else if (selectedIndex >= 3) {
    tft.fillRect(0, 180, 240, 60, ST77XX_WHITE);   
  }   

  if (selectedIndex >= 3) {
    tft.setTextColor(ST77XX_BLACK);
  } else {
    tft.setTextColor(ST77XX_WHITE);
  }
  tft.setTextWrap(true);
  tft.setTextSize(2);

  
  int listStartIndex = 0;
  if (selectedIndex > 3) {
    listStartIndex = selectedIndex - 3;
  }

  // check paired (bonded) devices with this remote controller
  int count = esp_bt_gap_get_bond_device_num();
  esp_err_t tError =  esp_bt_gap_get_bond_device_list(&count, pairedDeviceAddress);
  
  for (int i = listStartIndex; i < devicesCount; i++) {
    BTAdvertisedDevice* device = scanResult->getDevice(i);

    String deviceName = device->getName().c_str();
    if (deviceName.length() < 1) {
      deviceName = "No name";  //"no name";
    }
    esp_bd_addr_t* addr2 = device->getAddress().getNative();

    if (i == selectedIndex) {
       tft.setTextColor(ST77XX_BLACK);
       selectedDeviceName = deviceName;
//       esp_bd_addr_t* addr2 = device->getAddress().getNative();
       memcpy(selectedDeviceAddress, addr2, ESP_BD_ADDR_LEN);

       if (isPairedDevice(selectedDeviceAddress)) {
         deviceName = "(OK) " + deviceName;
       } else {
          deviceName = "(X) " + deviceName;
       }
    } else {
      tft.setTextColor(ST77XX_WHITE);
      uint8_t deviceAddr[6];
      memcpy(deviceAddr, addr2, ESP_BD_ADDR_LEN);
      if (isPairedDevice(deviceAddr)) {
        deviceName = "(OK) " + deviceName;
      } else {
        deviceName = "(X) " + deviceName;
      }
    }
   
    tft.setCursor(x, y);
    //tft.println(deviceName);
    tft.println(deviceName.substring(0, 19));
    y += 60;
  }

}

void displayConnectionMode() {

  // writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString)
  tft.fillScreen(ST77XX_BLACK);
  displayBitmap(50, 36, imageLogoColor, 48, 48);

  tft.fillCircle(140, 48, 5, 0x4208);
  writeText(160, 45, 1, ST77XX_WHITE, "MODE");

  tft.fillCircle(140, 64, 5, ST77XX_BLUE);
  writeText(160, 61, 1, ST77XX_WHITE, "SELECT");

  tft.fillRect(0, 119, 240, 1, ST77XX_WHITE);  
   
  if (useMenu) {
    tft.fillRect(0, 120, 240, 60, ST77XX_WHITE);
//    tft.setCursor(20, 140);
//    tft.setTextColor(ST77XX_BLACK);
//    tft.println("Master Mode");
    writeText(40, 140, 2, ST77XX_BLACK, "Master Mode");
//    tft.setCursor(40, 200);
//    tft.setTextColor(ST77XX_WHITE);
//    tft.println("Slave Mode");  
    writeText(40, 195, 2, ST77XX_WHITE, "Slave Mode");  
  } else {
//    tft.setCursor(20, 140);
//    tft.setTextColor(ST77XX_WHITE);
//    tft.println("Master Mode");
    writeText(40, 140, 2, ST77XX_WHITE, "Master Mode");
    tft.fillRect(0, 180, 240, 60, ST77XX_WHITE);   
//    tft.setCursor(20, 200);
//    tft.setTextColor(ST77XX_BLACK);
//    tft.println("Slave Mode");
    writeText(40, 195, 2, ST77XX_BLACK, "Slave Mode");  
  }
}

void displayMenuInformation(int kind) {
  // writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString)
  tft.fillScreen(ST77XX_BLACK);
  if (kind == 0) {
    // initializing search
     writeText(20, 100, 2, ST77XX_WHITE, "Searching...");
  } else if (kind == 1) {
    // remote device connected
     writeText(20, 100, 2, ST77XX_WHITE, "Connected.");
  } else if (kind == -1) {
    // failed to search remote devices
     writeText(20, 100, 2, ST77XX_WHITE, "Failed to search.");
  } else if (kind == 2) {
    // connecting...
    writeText(20, 100, 2, ST77XX_WHITE, "Connecting...");   
  } else if (kind == 3) {
    // failed to connecting...
    writeText(20, 100, 2, ST77XX_WHITE, "Failed to connect.");
  }
}

void scanBluetoothDevices() {
  if (CONNECTED) return;
  if (DEBUG) Serial.println("scanBluetoothDevices");

  selectedIndex = 0;
  displayMenuInformation(0);
  
  scanResult = SerialBT.discover(BT_DISCOVER_TIME);

  if (scanResult) {

    displayDevices();
    
  } else {

    displayMenuInformation(-1);
    
  }
}

void connectRemoteDevice() {
  if (CONNECTED) return;
  if (DEBUG) Serial.println("Connecting: " + selectedDeviceName);

  displayMenuInformation(2);

  bool connected = SerialBT.connect(selectedDeviceName);
  delay(1000);

  if (!connected) {
    connected = SerialBT.connect(selectedDeviceAddress);
    delay(1000);
  }

  CONNECTED = connected;
  if (connected) {
    Serial.println("Remote device connected.");
    CONNECTED = true;
    tft.fillScreen(ST77XX_BLACK);

    displayMenuInformation(1);

    delay(200);
    // send display type to remote camera
    if (displayOk) {
      sendDisplayTypeToRemote();
      delay(50);
      sendImageTypeToRemote();
    }

    delay(1000);
    
  } else {
    Serial.println("Failed to connect remote device.");

    displayMenuInformation(3);
    delay(4000);
    scanBluetoothDevices();
  }
}


void displayMenuInfo() {

  // Display (orange): scan remote bluetooth devices, or back to previous
  // Photo (red): list up
  // Video (green): list down
  // Action (blue): select remote device and connect

  if (CONNECTED) return;

  //writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString)
  tft.fillScreen(ST77XX_BLACK);
  displayBitmap(96, 32, imageLogoColor, 48, 48);

  tft.fillCircle(60, 110, 5, 0xFC4E);
  tft.fillCircle(60, 130, 5, ST77XX_RED);
  tft.fillCircle(60, 150, 5, ST77XX_GREEN);
  tft.fillCircle(60, 170, 5, ST77XX_BLUE);
  tft.fillCircle(60, 190, 5, 0x4208);
    
  writeText(100, 105, 1, ST77XX_WHITE, "Search");
  writeText(100, 125, 1, ST77XX_WHITE, "Scroll up");
  writeText(100, 145, 1, ST77XX_WHITE, "Scroll down");
  writeText(100, 165, 1, ST77XX_WHITE, "Connect");
  writeText(100, 185, 1, ST77XX_WHITE, "<< back");
  
}

// display Camera Remote Bluetooth logo image
void displayLogo() {
  if (!displayOk) return;

  if (DISPLAY_COLOR_MODE == 1) {
    // monochrome display
  } else if (DISPLAY_COLOR_MODE == 2) {                 


    if (menuState == 0) {
      // slave mode: camera connect remote controller

      tft.fillScreen(ST77XX_BLACK);
      //displayBitmap(72, 72, imageLogoColor, 96, 96);
      displayBitmap(96, 48, imageLogoColor, 48, 48);     
      writeText(60, 130, 1, ST77XX_WHITE, "Connect from camera");
//      writeText(100, 160, 1, ST77XX_WHITE, "from");
//      writeText(90, 190, 1, ST77XX_WHITE, "camera");

      tft.fillCircle(70, 160, 5, 0x4208);
      writeText(100, 155, 1, ST77XX_WHITE, "<< back");      

    } else if (menuState == 1) {
      // master / slave selection menu

      displayConnectionMode();

    } else if (menuState == 2) {
      // master mode menu screen

      displayMenuInfo();   
            
    }

    delay(500);
  }
}

// Bluetooth connection status call back
void callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if(event == ESP_SPP_SRV_OPEN_EVT){
    Serial.println("Client Connected");
    //digitalWrite(LED,HIGH);
    CONNECTED = true;
    tft.fillScreen(ST77XX_BLACK);

    // send display type to remote camera
    if (displayOk) {
      sendDisplayTypeToRemote();
      delay(50);
      sendImageTypeToRemote();
    }
  }
 
  if(event == ESP_SPP_CLOSE_EVT ){
    Serial.println("Client disconnected");
    //digitalWrite(LED,LOW);
    CONNECTED = false;
    tft.fillScreen(ST77XX_BLACK);
    displayLogo();
  }
}

void resetBluetooth() {
    if (CONNECTED) {
      SerialBT.flush();
      SerialBT.disconnect();
    }
    SerialBT.end();
    delay(500);

    // restart bluetooth
    SerialBT.begin("Camera Remote Bluetooth", useMenu);    //Bluetooth device name 
    SerialBT.register_callback(callback);

    delay(500);  
}

// memory allocation for image frame data receiving from remote camera
// memory is prepared on setup()
void prepareImageDataReceiving() {
  heap_caps_print_heap_info(MALLOC_CAP_DEFAULT);

//  Serial.println(ESP.getHeapSize());
//  Serial.println(ESP.getFreeHeap());
//  Serial.println(ESP.getPsramSize());
//  Serial.println(ESP.getFreePsram());
  
//  heap_caps_malloc(12288 * sizeof(uint16_t) * 2, MALLOC_CAP_SPIRAM);
//  image_data_receiving = (uint16_t*)ps_malloc(12288 * sizeof(uint16_t));

  image_data_receiving = (byte*)malloc(12288 * sizeof(byte));

  if (!image_data_receiving) {
    Serial.println("failed for image_data_receiving malloc");
  } else {
    Serial.println("success for image_data_receiving malloc");
  }
}

// display specific icon bitmaps depends on the current camera status
void displayIcons(bool photo, bool video, bool recording, bool flash) {
  if (photo) {
    clearArea(30, 180, 40, 40, ST77XX_BLACK);
    displayBitmap(30, 180, imageIconPhotoB, 40, 40);
  } else if (!photo) {
    clearArea(30, 180, 40, 40, ST77XX_BLACK);
    displayBitmap(30, 180, imageIconPhoto, 40, 40);
  }
  if (recording) {
    clearArea(100, 180, 40, 40, ST77XX_BLACK);
    displayBitmap(100, 180, imageIconVideoRecording, 40, 40);
  } else {
    if (video) {
      clearArea(100, 180, 40, 40, ST77XX_BLACK);
      displayBitmap(100, 180, imageIconVideoB, 40, 40);
    } else if (!video) {
      clearArea(100, 180, 40, 40, ST77XX_BLACK);
      displayBitmap(100, 180, imageIconVideo, 40, 40);
    }
  }
  if (flash) {
    clearArea(170, 180, 40, 40, ST77XX_BLACK);
    displayBitmap(170, 180, imageIconFlash, 40, 40);
  } else if (!flash) {
    clearArea(170, 180, 40, 40, ST77XX_BLACK);
    displayBitmap(170, 180, imageIconFlashNone, 40, 40);
  }
}


// display icons and text information depends on the status
// display mode: none, text, image 
void displayFrame(int frameKind) {
  if (!displayOk) return;

// ST7789 *************************************************
  if (frameKind < 0) {
    if (MODE_DISPLAY == 0) {
        // none
        displayIcons(false, false, false, false);
        currentDisplayMode = 0;
      } else if (MODE_DISPLAY == 1) {
        // photo mode
        displayIcons(true, false, false, false);
        currentDisplayMode = 1;
      } else if (MODE_DISPLAY == 2 || MODE_DISPLAY == 3) {
        // video mode
        displayIcons(false, true, false, false);
        currentDisplayMode = 2;
      }
  } else {
    if (frameKind == 0) {
      // none
      displayIcons(false, false, false, MODE_FLASH);
      currentDisplayMode = 0;
    } else if (frameKind == 1) {
      // photo mode
      displayIcons(true, false, false, MODE_FLASH);
      currentDisplayMode = 1;
    } else if (frameKind == 2) {
      // video mode
      displayIcons(false, true, false, MODE_FLASH);
      currentDisplayMode = 2;
    } else if (frameKind == 3) {
      // video recording mode
      displayIcons(false, false, true, MODE_FLASH);
      currentDisplayMode = 3;
    }
    if (MODE_DISPLAY == 0) {
      //writeText(uint16_t x, uint16_t y, uint16_t textSize, uint16_t textColor, String textString)
      clearArea(0, 0, 240, 160, ST77XX_BLACK);
      if (currentDisplayMode == 0) {
        writeText(60, 80, 4, ST77XX_WHITE, "READY");
      } else if (currentDisplayMode == 1) {
        writeText(60, 80, 4, ST77XX_WHITE, "PHOTO");
      } else if (currentDisplayMode == 2) {
        writeText(60, 80, 4, ST77XX_WHITE, "VIDEO");
      } else if (currentDisplayMode == 3) {
        if (MODE_VIDEO_RECORDING) {
          writeText(60, 80, 4, ST77XX_WHITE, "REC.");
        } else {
          writeText(60, 80, 4, ST77XX_WHITE, "REC..");
        }
      }
    }
  }

}

// resize image frame data (96 x 64) into full size (192 x 128)
// reduces data transfer time and memory consumption
// This version does not use
void resizeImage(String colorMode) {
  // for the best display result (faster performance)
  // use only small size data frame (96x64) of 256 color image
  // and resize the image at runtime for displaying

  // 96 x 64 >> 192 x 128
  // pixels: 6144 >> 24576 (4 times of original pixels)
  // src.W: 96, src.H: 64, dst.W: 192, dst.H: 128
  // src.Pixel >> dst.Pixels (4 pixels)
  // src.PixelIndex: (src.X + (src.Y * src.W))
  // dst.PixelIndex1: (src.X * 2) + ((src.Y * 2) * dst.W)
  // dst.PixelIndex2: ((src.X * 2) + 1) + ((src.Y * 2) * dst.W)
  // dst.PixelIndex3: (src.X * 2) + (((src.Y * 2) + 1) * dst.W)
  // dst.PixelIndex4: ((src.X * 2) + 1) + (((src.Y * 2) + 1) * dst.W)

  int srcW = 96, srcH = 64;
  int dstW = 192, dstH = 128;
  uint16_t srcPixel = 0;

  if (DEBUG) {
    Serial.println("Resizing frame data started.");
  }

  if (colorMode.equalsIgnoreCase("8bit")) {
    // 8bit 256 color
    for(int j = 0; j < srcH; j++) {
      for(int i = 0; i < srcW; i++) {
        srcPixel = ByteToInteger16(image_data_receiving[i + (j * srcW)]);
        image_data_color[(i * 2) + ((j * 2) * dstW)] = srcPixel;
        image_data_color[((i * 2) + 1) + ((j * 2) * dstW)] = srcPixel;
        image_data_color[(i * 2) + (((j * 2) + 1) * dstW)] = srcPixel;
        image_data_color[((i * 2) + 1) + (((j * 2) + 1) * dstW)] = srcPixel;
      }
    }
  } else if (colorMode.equalsIgnoreCase("16bit")) {
    // 16bit 65k color 
    int dstX = 0;
    for(int j = 0; j < srcH; j++) {
      for(int i = 0; i < srcW * 2; i+=2) {
        ArrayToInteger16 intData;
        intData.barry[0] = image_data_receiving[i + (j * srcW * 2)];
        intData.barry[1] = image_data_receiving[i + 1 + (j * srcW * 2)];
        srcPixel = intData.integer16;
        
        dstX = (i / 2);
        image_data_color[(dstX * 2) + ((j * 2) * dstW)] = srcPixel;
        image_data_color[((dstX * 2) + 1) + ((j * 2) * dstW)] = srcPixel;
        image_data_color[(dstX * 2) + (((j * 2) + 1) * dstW)] = srcPixel;
        image_data_color[((dstX * 2) + 1) + (((j * 2) + 1) * dstW)] = srcPixel;
      }
    }    
  }
  tft.drawRGBBitmap(24, 24, image_data_color, 192, 128);    
}

// Jpeg decoding and display jpeg image to display
byte* jpegData;
void displayJpeg(int dataLength) {
  //byte* jpegData;
  jpegData = (byte*)malloc(dataLength * sizeof(byte));

  for(int i = 0; i < dataLength; i++) {
    jpegData[i] = image_data_receiving[i];
  }

  TJpgDec.drawJpg(24, 24, jpegData, dataLength * sizeof(byte));

  free(jpegData);
}

// display bitmap image: remote frame data
// display mode: image
void displayImage() {
  if (!displayOk) return;

  if (DEBUG) {
    Serial.println("Displaying bitmap...");
  }
  //clearArea(0, 0, 240, 160, ST77XX_BLACK);
  if (resizeFrame) {
    displayBitmap(24, 24, image_data_color, 192, 128);  
  } else {
    displayBitmap(72, 44, image_data_color, 96, 64);
  }

}

// display remote camera status
// display mode: text
void displayInfo(String cameraMode, String battery, String storage, String sdcard) {
  if (!displayOk) return;

// ST7789 ***************************************************************
  if (cameraMode.equalsIgnoreCase("photo")) {
    displayIcons(true, false, false, MODE_FLASH);
  } else if (cameraMode.equalsIgnoreCase("video")) {
    displayIcons(false, true, false, MODE_FLASH);
  } else if (cameraMode.equalsIgnoreCase("recording")) {
    displayIcons(false, false, true, MODE_FLASH);
  }

  clearArea(0, 0, 240, 160, ST77XX_BLACK);
  writeText(40, 40, 2, ST77XX_WHITE, "BATTERY: " + battery + " %");
  writeText(40, 80, 2, ST77XX_WHITE, "DISK: " + storage);
  writeText(40, 120, 2, ST77XX_WHITE, "SD: " + sdcard);
}

// clears the data value array
void clearDataValues() {
  int numberOfValues = 10;    //(int)sizeof(dataValues)/sizeof("hello");
  for (int i = 0; i < numberOfValues; i++) {
    dataValues[i] = "";
  }
}

// parse the remote data headers and values
// split data string into array
// delimeter: ":"
void parseData(String dataString) {
  clearDataValues();

  // data delimeter: ":"
  char delimeter = ':';
  int findStart = 0;
  int findIndex = 0;
  int findCount = 0;

  findIndex = dataString.indexOf(delimeter, findStart);
  if (findIndex > 0) {
    dataValues[findCount] = dataString.substring(findStart, findIndex);
    findCount ++;
    findStart = findIndex + 1;

    while (findIndex > 0) {
      findIndex = dataString.indexOf(delimeter, findStart);
      if (findIndex > 0) {
        dataValues[findCount] = dataString.substring(findStart, findIndex);
        findCount ++;
        findStart = findIndex + 1;        
      }
    } 
  }
}

// check the buttons (switches) status
// and sending command to the remote camera
void processAction() {

  if (APP_MODE == 1) {
    // Joystick Mode: not used 
    //    Serial.println("Joystick mode");
  } else if (APP_MODE == 0) {
    // Switches Mode
    //    Serial.println("Switches Mode");

    int photo = digitalRead(PIN_BUTTON_PHOTO);
    int video = digitalRead(PIN_BUTTON_VIDEO);
    int flash = digitalRead(PIN_BUTTON_FLASH);
    // Action Button for both modes
    int action = digitalRead(BUTTON_ACTION);
    // Change Camera
    int camera = digitalRead(BUTTON_CHANGE_CAMERA);
    // Display Mode: toggle to (lcd image >> lcd text >> no display)
    int displayMode = digitalRead(BUTTON_DISPLAY_MODE);


    if (!CONNECTED) {
    // Menu button actions

      //Serial.println("menuState: " + String(menuState) + ", camera: " + String(camera) + ", action: " + String(action));

      if (menuState == 0) {
        // slave mode screen
        if (!camera) {
          menuState = 1;
          displayLogo();          
        }
      } else if (menuState == 1) {
        // App Slash Screen
        // Select master or slave mode screen

        if (!camera) {
          // change master/slave mode menu
          useMenu = !useMenu;
          
          //displayConnectionMode();
          displayLogo();

        } else if (!action) {
          // select connection mode
          // if master mode, show master mode menu
          // useMenu == true: master mode
          // useMenu == false: slave mode
          menuState = (useMenu ? 2 : 0);
          displayLogo();

          resetBluetooth();
        }

      } else if (menuState == 2) {
        // master mode menu screen
       
        if (!displayMode) {
          // searching the remote devices
          if (DEBUG) Serial.println("photo button...");
          scanBluetoothDevices();
        } else if (!photo) {
          // device list select up
          if (selectedIndex > 0) {
            selectedIndex -= 1;
            displayDevices();
          }
        } else if (!video) {
          // device list select down
          if (selectedIndex < devicesCount - 1) {
            selectedIndex += 1;
            displayDevices();
          }
        } else if (!action) {
          // connect remote device
          connectRemoteDevice();
        } else if (!camera) {
          //back to menu
//          displayLogo();
          menuState = 1;
          displayLogo();

        }
  
      }
    
      ACTION_STATUS = 1;        
 
      delay(200);
      return;
    }




    if (ACTION_STATUS == 0) {
      
      if (!flash) {
        MODE_FLASH = !MODE_FLASH;
      } else if (!action) {
        // do nothing
        if (isVideoRecording) {
          isVideoRecording = false;
        }
      } else {
        // values from push button
      }

      if (!displayMode) {
        if (MODE_DISPLAY == 2) {
          MODE_DISPLAY = 0;
        } else if (MODE_DISPLAY == 1) {
          MODE_DISPLAY = 2;
        } else if (MODE_DISPLAY == 0) {
          MODE_DISPLAY = 1;
        }
        // clear display area only the display mode has changed, once...
        clearArea(0, 0, 240, 160, ST77XX_BLACK);
      }
  
      if (!photo || !video || !flash || !action || !camera || !displayMode) {

        tft.fillScreen(ST77XX_BLACK);
  
        String actionString = "iot:";
        String displayType = "none";
        if (DISPLAY_COLOR_MODE == 0) {
          displayType = "none";
        } else if (DISPLAY_COLOR_MODE == 1) {
          displayType = "SD1306";
        } else if (DISPLAY_COLOR_MODE == 2) {
          displayType = "ST7789";
        }

        if (!camera || !displayMode) {
          actionString += String(MODE_PHOTO) + ":" 
              + String(MODE_VIDEO) + ":" 
              + String(MODE_FLASH) + ":" 
              + String(!action) + ":" 
              + String(!camera) + ":" 
              + String(!displayMode) + ":"
              + displayType;      
        } else {
          actionString += String(!photo) + ":" 
              + String(!video) + ":" 
              + String(MODE_FLASH) + ":" 
              + String(!action) + ":" 
              + String(!camera) + ":" 
              + String(!displayMode) + ":"
              + displayType;
        }


        if (CONNECTED) {

          SerialBT.println(actionString);
          SerialBT.write(13);
          SerialBT.write(10);
      
          ACTION_STATUS = 1;

          if (DEBUG) {
            Serial.println(actionString);          
          }
        }

        delay(100);     
      }            
    }

    if (ACTION_STATUS == 1) {

      if (photo && video && flash && action && camera && displayMode) {
        ACTION_STATUS = 0;
        delay(10);      
      }
    
    }
  } 
}

// analize the remote frame data header
// process and convert into bitmap from remote frame data bytes
void processImageData(String header) {
  // image data header received: "imagedata:xxxxx:16bit" image data pixel count, and color mode (16 bit or 8 bit)
  // new image format: "jpeg"
  
  if (DEBUG) {
    Serial.println(header);
  }

  // flag data processing, prevent another incoming data processing
  // until this image display session finished
  imageDataProcessing = true;
  
  while (imageDataProcessing) {

    long now = millis();
//    if (endTime - now <= 0) {
//      imageDataProcessing = false;
//      break;
//    }

    if (MODE_DISPLAY != 2) {
      imageDataProcessing = false;
      break;
    }

    // image frame data information header to receive into memory
    String pixelsDataInfo = SerialBT.readStringUntil('\n');

    // parse data string into values
    parseData(pixelsDataInfo);

    String dataName = dataValues[0];
    String colorMode = dataValues[4];
    int totalPixels = dataValues[2].toInt();
           
    if (dataName.startsWith("imagepixels_color")) {

      if (colorMode.startsWith("8bit")) {
        // 256 color frame data
        
        //memset(image_data_receiving, 0, sizeof(image_data_receiving));
        //memset(image_data_color, 0, sizeof(image_data_color));

        int count = 0;
        while (count < totalPixels) {
            count += SerialBT.readBytes(image_data_receiving, totalPixels);                       
        }
        if (count <= 0) {
          delay(100);
        } else {
          //delay(20); 
          if (resizeFrame) {
            resizeImage(colorMode);   
          }
        }      
        imageDataProcessing = false;  
      } else if (colorMode.startsWith("16bit")) {
        // 65k color frame data
        
        int count = 0;
        while (count < totalPixels * 2) {
            count += SerialBT.readBytes(image_data_receiving, totalPixels * 2);
        }
        //Serial.println("Total Read count: " + String(count));
        if (count <= 0) {
          delay(100);
        } else {
          //delay(20);
          if (resizeFrame) {
            resizeImage(colorMode);   
          }
        }             
        imageDataProcessing = false;  
      } else if (colorMode.startsWith("jpeg")) {
        // jpeg decode method
        
        if (DEBUG) {
          Serial.println(pixelsDataInfo);
        }
        
        int count = 0;
        while (count < totalPixels) {
          count += SerialBT.readBytes(image_data_receiving, totalPixels);
        }

        if (count <= 0) {
          delay(100);
        } else {
          displayJpeg(totalPixels);        
        }
        imageDataProcessing = false;
      }

    } else {  // if (pixelsDataInfo.equalsIgnoreCase("imagepixels_color")
      if (DEBUG) {
        Serial.println("Unexpected data: " + pixelsDataInfo);     
      }
      imageDataProcessing = false;
    }

    // image frame information header: analize for remote camera status
    // display proper icon images accordingly
    parseData(header);
  
    int pixels = dataValues[1].toInt();
    colorMode = dataValues[2];     // 16bit: 65k 16 bit color mode, 8bit: 256 8 bit color mode, jpeg format (this version is using)
    int photoMode = dataValues[3].toInt();
    int videoMode = dataValues[4].toInt();
    int flashMode = dataValues[5].toInt();
    int displayMode = dataValues[6].toInt();
    int previewSize = dataValues[7].toInt();
    int dataSize = dataValues[8].toInt();
  
    //bool previewSizeChanged = false;
    if (MODE_PREVIEW_SIZE != previewSize) {
      // the preview size has changed, clear the screen
      //clearArea(0, 0, 240, 160, ST77XX_BLACK);
    }
  
    pixelsPerDataReceiving = dataSize;
    MODE_PREVIEW_SIZE = previewSize;    
    MODE_PHOTO = photoMode;
    MODE_VIDEO = (videoMode == 0 ? 0 : 1);
    MODE_VIDEO_RECORDING = (videoMode == 2 ? 1 : 0);
    MODE_FLASH = flashMode;
    MODE_COLOR = colorMode;
    MODE_DISPLAY = displayMode;
  
    // process status icon display
    displayIcons(MODE_PHOTO, MODE_VIDEO, MODE_VIDEO_RECORDING, MODE_FLASH);

    // check and process the user button actions
    processAction();
  } // end of while(imageDataProcessing) 
          
}

// analize and process data string received from remote camera
void processData(String dataString) {

  // if display mode is Image Mode and data receiving, do not process here to prevent interruption.
  if (imageDataProcessing) {
    return;
  }
  
  if (!dataString.startsWith("imagepixels")) {
    if (DEBUG) {
      Serial.println("processData: " + dataString);      
    }
  }

  // parse data string into values
  // split string into array
  // delemiter: ":"
  parseData(dataString);

  // process parsed data values
  // data name: dataValues[0], 'photo', 'video', 'iot', 'imagedata', 'imagefixels', 'received', , 'imagefixels_color'

  String dataName = dataValues[0];
  
  if (dataName.equalsIgnoreCase("iot") || dataName.equalsIgnoreCase("received")) {
    String statusPhoto = dataValues[1];
    int photoValue = statusPhoto.toInt();
    String statusVideo = dataValues[2];
    int videoValue = statusVideo.toInt();
    String statusFlash = dataValues[3];
    int flashValue = statusFlash.toInt();
    String statusAction = dataValues[4];
    int actionValue = statusAction.toInt();
    String statusResult = dataValues[5];
    int resultValue = statusResult.toInt();

    String statusDisplay = "0";
    // display mode: none, text, image
    int numberOfElements = ARRAY_SIZE(dataValues);

    statusDisplay = dataValues[6];  // none, text, image
    int displayValue = statusDisplay.toInt();
    if (CONNECTED) {
      //Serial.println("MODE_DISPLAY: " + String(MODE_DISPLAY) + ", displayValue: " + String(displayValue));
      if (!dataName.equalsIgnoreCase("received")) {
        if (MODE_DISPLAY != displayValue) {
          MODE_DISPLAY = displayValue; 
          // display mode changed, clear the display screen
          clearArea(0, 0, 240, 160, ST77XX_BLACK);
        }
      }

      if (DEBUG) {
        Serial.println("MODE_DISPLAY: " + String(MODE_DISPLAY));
      }
    } else {
        // would not go this codes
        // Display logo image        
        displayLogo();        
    }

    String previewSize = "0";
    previewSize = dataValues[7];
    int previewSizeValue = previewSize.toInt();
    MODE_PREVIEW_SIZE = previewSizeValue;    

    String pixelsOnData = String(pixelsPerDataReceiving);
    pixelsOnData = dataValues[8];
    pixelsPerDataReceiving = pixelsOnData.toInt();

    MODE_PHOTO = photoValue;
    MODE_VIDEO = videoValue;

    // for LED blinking status
    MODE_VIDEO_RECORDING = MODE_VIDEO;
   
    MODE_FLASH = flashValue;

    digitalWrite(PIN_LED_PHOTO, MODE_PHOTO);
    digitalWrite(PIN_LED_VIDEO, MODE_VIDEO);

    if (displayOk) {
      if (MODE_DISPLAY == 0) {
        if (MODE_PHOTO) {
          displayFrame(1);
        } else if (MODE_VIDEO) {
          displayFrame(2);
        }
      } else if (MODE_DISPLAY == 2) {
        // For ST7789, it is image display mode then clear area
        //clearArea(0, 0, 240, 160, ST77XX_BLACK);
        if (MODE_PHOTO) {
          displayFrame(1);
        } else if (MODE_VIDEO) {
          if (isVideoRecording) {
            displayFrame(3);
          } else {
            displayFrame(2);
          }
        }        
      } else if (MODE_DISPLAY == 3) {
        if (MODE_PHOTO) {
          displayFrame(1);
        } else if (MODE_VIDEO) {
          if (!isVideoRecording) {
            displayFrame(2);
          }
        }                
      }
    }    
  } else if (dataName.equalsIgnoreCase("info")) {
    // display mode: text

    String cameraMode = dataValues[1];
    String batteryLevel = dataValues[2];
    String storage = dataValues[3];
    String sdcard = dataValues[4];

    if (displayOk) {
      displayInfo(cameraMode, batteryLevel, storage, sdcard);    
    }
  } else if (dataName.equalsIgnoreCase("photo")) {
    String result = dataValues[1];
    int resultValue = result.toInt();
    if (resultValue == 1) {
      digitalWrite(PIN_LED_PHOTO, HIGH);
    }
    displayFrame(1);  
    
    // for LCD icon
    isVideoRecording = false;
       
  } else if (dataName.equalsIgnoreCase("video")) {
    String result = dataValues[1];
    int resultValue = result.toInt();
    if (resultValue == 1) {
      // for LCD icon
      isVideoRecording = false;
      
      digitalWrite(PIN_LED_VIDEO, HIGH); 
      displayFrame(2);       
    } else if (resultValue == 2) {
      // recording, blink (toggle) video status led
      MODE_VIDEO_RECORDING = !MODE_VIDEO_RECORDING;
      digitalWrite(PIN_LED_VIDEO, MODE_VIDEO_RECORDING);

      // for LCD icon
      isVideoRecording = true;
      
      if (MODE_DISPLAY == 0) {
        displayFrame(3);        
      } else if (MODE_DISPLAY == 2) {
        displayFrame(3);
      }
    }
  } else if (dataName.equalsIgnoreCase("imagepixels")){
    // SD1306 *******************************************************
  } else if (dataName.equalsIgnoreCase("imagedata")) {
    // image frame header, should not run here
    // backup processing for error while processing the processImageData function
    MODE_COLOR = dataValues[2];
    processImageData(dataString);
  }
}
