/**********************************************************************
  Filename    : variables.h
  Description : variables, constant, and definition members
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/09/09
**********************************************************************/

// SSD1306 OLED Display: 128x64 monochrome (black and white)
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

#define MCU_MODE 1                // 1: ESP32 WROOM 32, 2: ESP32 WROVER
#define DISPLAY_COLOR_MODE 2      // 0: no display, 1: Monochrome (SD1306, black and white), 2: Color (ST7789)

// ESP32 WROOM D1 MINI: SPI Connection *****************************************
// SPI: VSPI (serial peripheral interface)
// With Adafruit library, the ST7789 need to use Hardware SPI calling. 
// Default hardware pins: MOSI (23), CLK (18), CS (5)
// Ueer selective pins: DC, RESET
// ST7789 : SPI (VSPI) ***************************************************************
#define TFT_CS         5      // set pin number for library calling, but do not connect to mcu for ST7789 display and SPI mode (may need to connet for other display) 
#define TFT_RST        4      // any available pin
#define TFT_DC         16     // any available pin
#define TFT_MOSI       23     // Data out
#define TFT_SCLK       18     // Clock out
// For 1.14", 1.3", 1.54", 1.69", and 2.0" TFT with ST7789:
// Hardware SPI
// Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);
// Software SPI
// Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);
   
// replace this statement with other display for correct display and versions
Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);

// not used implemenation for this version
//#define JX_PHOTO_VIDEO      13          // Joystick X:  photo and video mode
//#define JY_FLASH            14          // Joystick Y:  flash on / off
//#define PIN_APP_MODE        21          // app mode: joystick or buttons mode
int APP_MODE = 0;                         //0: Switched Mode, 1: Joystick Mode

#define PIN_BUTTON_PHOTO    25            // photo mode
#define PIN_BUTTON_VIDEO    27            // video mode
#define PIN_BUTTON_FLASH    33            // flash mode

#define BUTTON_ACTION           32        // Action Button: camera action: either take photo or record video according to current mode
#define BUTTON_CHANGE_CAMERA    26        // change camera: front or back cameras

#define BUTTON_DISPLAY_MODE   17          // WROVER: 12           // display mode: Image, Text, No Display
#define PIN_LED_PHOTO         19          //23                    // photo mode led (not used)
#define PIN_LED_VIDEO         35          //22                    // video mode led (not used)

#define ARRAY_SIZE(x) sizeof(x)/sizeof(x[0])

// 192 x 128: image frame data buffer for displaying
uint16_t image_data_color[24576];
// 96 x 64
//uint16_t PROGMEM image_data_color[6144];

byte* image_data_receiving;               // temporary frame data buffer for remote image

BluetoothSerial SerialBT;

bool DEBUG = true;                       // debug information display in serial monitor
int CONNECTED = 0;                        // remote bluetooth connection status

bool displayOk = false;                   // check if display is connected or working condition
long modeChangedTime = -1;
long currentTime;
bool resizeFrame = true;                  // current version: remote frame image resizing for full display
bool imageJpeg = true;

int ACTION_STATUS = 0;                    // 0: nothing, 1: action started, 2: action(video) processing
int MODE_PHOTO = 1;                       // 1: photo mode, 0: video mode
int MODE_VIDEO = 0;                       // 1: video mode, 0: photo mode
int MODE_VIDEO_RECORDING = 0;             // recording video led flashing
int MODE_FLASH = 0;                       // 1: torch flash, 0: auto
int FLASH_STATUS = 0;                     // 1: torch flash, 0: auto

int MODE_DISPLAY = 0;                     // 0: no display, 1: text display, 2: image display
int MODE_PREVIEW_SIZE = 1;                // 0: 96x64, 1: 128x96, 2: 192x128
String MODE_COLOR = "16bit";
bool imageDataProcessing = false;         // on/off flag for frame data processing and displaying
int currentDisplayMode = 0;               // current display mode
bool isVideoRecording = false;            // current video mode status: recording or not

String dataValues[15];                    // array of string values: remote data, status, frame values
char cphoto, cvideo, cflash, caction, cresult, ccamera;     // remote camera and action status  
int iphoto, ivideo, iflash, iaction, iresult, icamera;      // remote camera and action status converted integer
String sreturn;
int pixelsPerDataReceiving = 192;         // the data bytes size from remote camera frame: not used in this version

// Menu Implemenataion
int menuState = 1;      // Slave mode state: 0, Master/Slave selection menu screen state: 1, Master menu screen state: 2, 
bool useMenu = false;   // true: Master mode (connect camera from remote controller), false: Slave mode (connect remote controller from camera)
#define BT_DISCOVER_TIME  10000     // 10 seconds to scan remote cameras
int devicesCount = 0;
int selectedIndex = 0;
String selectedDeviceName = "";
uint8_t selectedDeviceAddress[6];
BTScanResults *scanResult;
uint8_t addressToConnect[6];        // remote bluetooth device address
String strAddressToConnect;         // remote bluetooth device address string representation
String nameToConnect;               // remote bluetooth device name
#define PAIR_MAX_DEVICES 20
uint8_t pairedDeviceAddress[PAIR_MAX_DEVICES][6];
